-- camera.lua :: a basic camera adapted to the asset/world system.
-- Use hump.camera as the view backend.

--[[
  Copyright © 2019 Kazhnuz

  Permission is hereby granted, free of charge, to any person obtaining a copy of
  this software and associated documentation files (the "Software"), to deal in
  the Software without restriction, including without limitation the rights to
  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
  the Software, and to permit persons to whom the Software is furnished to do so,
  subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]

local cwd   = (...):gsub('%.camera$', '') .. "."

local CameraSystem  = Object:extend()
local View          = require(cwd .. "libs.hump.camera")

local SPLITSCREEN_ISVERTICAL  = false
local SCREEN_LIMIT            = 4

-- INIT FUNCTIONS
-- Initialize the camera system

function CameraSystem:new(world)
  self.scene = world.scene
  self.world = world

  self.verticalSplit = SPLITSCREEN_ISVERTICAL

  self:initViews()
end

function CameraSystem:initViews()
  self.views = {}

  self.views.list   = {}
  self.views.basewidth, self.views.baseheight   = core.screen:getDimensions()
  self.views.width,     self.views.height       = self:getViewsDimensions()

  self.views.posList = {}
  self.views.posList.dual   = {}
  self.views.posList.multi  = {}

  if (self.verticalSplit) then
    self.views.posList.dual[1] = {}
    self.views.posList.dual[1].x = 0
    self.views.posList.dual[1].y = (self.views.baseheight/4)

    self.views.posList.dual[2] = {}
    self.views.posList.dual[2].x = 0
    self.views.posList.dual[2].y = -(self.views.baseheight/4)
  else
    self.views.posList.dual[1] = {}
    self.views.posList.dual[1].x = -(self.views.basewidth/4)
    self.views.posList.dual[1].y = 0

    self.views.posList.dual[2] = {}
    self.views.posList.dual[2].x = (self.views.basewidth/4)
    self.views.posList.dual[2].y = 0
  end

  self.views.posList.multi[1] = {}
  self.views.posList.multi[1].x = -(self.views.basewidth /4)
  self.views.posList.multi[1].y =  (self.views.baseheight/4)

  self.views.posList.multi[2] = {}
  self.views.posList.multi[2].x =  (self.views.basewidth /4)
  self.views.posList.multi[2].y =  (self.views.baseheight/4)

  self.views.posList.multi[3] = {}
  self.views.posList.multi[3].x = -(self.views.basewidth /4)
  self.views.posList.multi[3].y = -(self.views.baseheight/4)

  self.views.posList.multi[4] = {}
  self.views.posList.multi[4].x = (self.views.basewidth /4)
  self.views.posList.multi[4].y = -(self.views.baseheight/4)
end

-- INFO FUNCTIONS
-- Get informations from the camera system

function CameraSystem:getViewNumber()
  return #self.views.list
end

function CameraSystem:haveView()
  return (self:getViewNumber() == 0)
end

function CameraSystem:getViewsDimensions(viewNumber)
  local basewidth, baseheight = self.views.basewidth, self.views.baseheight
  local viewnumber = viewNumber or self:getViewNumber()

  if (viewnumber <= 1) then
    return basewidth,       baseheight
  elseif (viewnumber == 2) then
    if (self.verticalSplit) then
      return (basewidth),     (baseheight/2)
    else
      return (basewidth/2),   (baseheight)
    end
  else
    return (basewidth/2),   (baseheight/2)
  end
end

function CameraSystem:recalculateViewsPositions()
  if #self.views.list == 1 then
    self.views.list[1].pos.onScreen.x = 0
    self.views.list[1].pos.onScreen.y = 0
  else
    for i,v in ipairs(self.views.list) do
      local x, y = self:getViewPositions(i)
      self.views.list[i].pos.onScreen.x = x
      self.views.list[i].pos.onScreen.y = y
    end
  end
end

function CameraSystem:getViewPositions(id)
  local viewNumber = #self.views.list

  if (viewNumber == 2) and ((id == 1) or (id == 2)) then
    return self.views.posList.dual[id].x,   self.views.posList.dual[id].y
  elseif (viewNumber > 2) then
    return self.views.posList.multi[id].x,  self.views.posList.multi[id].y
  end

end

-- WRAPPER and UTILS
-- Access data from the views

function CameraSystem:addView(x, y, target)
  if (#self.views.list < SCREEN_LIMIT) then
    local id = #self.views.list + 1
    local view = {}

    view.pos = {}
    view.pos.x = x or 0
    view.pos.y = y or 0
    view.pos.onScreen = {}

    view.cam    = View(view.pos.x, view.pos.y, 1, 0, true)
    -- TODO: add a target system in order to make a camera able
    -- to target a specific object
    view.target = target

    table.insert(self.views.list, view)
    self.views.width, self.views.height = self:getViewsDimensions()
    self:recalculateViewsPositions()
  end
end

function CameraSystem:getView(id)
  return self.views.list[id]
end

function CameraSystem:getViewCam(id)
  local view = self:getView(id)

  return view.cam
end

function CameraSystem:attachView(id)
  if (id ~= nil) then
    local cam = self:getViewCam(id)

    local viewx, viewy, vieww, viewh = self:getOnScreenViewCoordinate(id)
    cam:attach()
    love.graphics.setScissor(viewx, viewy, vieww, viewh)
  end
end

function CameraSystem:detachView(id)
  if (id ~= nil) then
    local cam = self:getViewCam(id)

    love.graphics.setScissor( )
    cam:detach()
  end
end

function CameraSystem:getViewCoordinate(id)
  local cam = self:getViewCam(id)

  local camx, camy, camw, camh
  camx = cam.x - (self.views.width/2)
  camy = cam.y - (self.views.height/2)

  camw = self.views.width
  camh = self.views.height
  return camx, camy, camw, camh
end

function CameraSystem:getOnScreenViewCoordinate(id)
  local view = self:getView(id)

  local viewx, viewy, vieww, viewh
  local basex, basey = (self.views.basewidth  / 2), (self.views.baseheight / 2)
  viewx   = (basex) + view.pos.onScreen.x - (self.views.width   / 2)
  viewy   = (basey) + view.pos.onScreen.y - (self.views.height  / 2)

  vieww   = self.views.width
  viewh   = self.views.height
  return viewx, viewy, vieww, viewh
end

function CameraSystem:getOnScreenViewRelativePosition(id)
  local view = self:getView(id)

  local viewx, viewy
  local basex, basey = (self.views.basewidth  / 2), (self.views.baseheight / 2)
  viewx   = view.pos.onScreen.x
  viewy   = view.pos.onScreen.y

  return viewx, viewy
end

function CameraSystem:getOnScreenViewCenter(id)
  local view = self:getView(id)

  local viewx, viewy
  local basex, basey = (self.views.basewidth  / 2), (self.views.baseheight / 2)
  viewx   = (basex) + view.pos.onScreen.x
  viewy   = (basey) + view.pos.onScreen.y

  return viewx, viewy
end

function CameraSystem:getViewScale(id)
  local cam = self:getViewCam(id)

  return cam.scale
end

function CameraSystem:limitView(id)
  local viewx, viewy, vieww, viewh  = self:getViewCoordinate(id)
  local worldw, worldh              = self.world:getDimensions()
  local posx = self.views.list[id].pos.x
  local posy = self.views.list[id].pos.y
  local minx = self.views.width  / 2
  local miny = self.views.height / 2
  local maxx = worldw - minx
  local maxy = worldh - miny

  self.views.list[id].pos.x = utils.math.between(posx, minx, maxx)
  self.views.list[id].pos.y = utils.math.between(posy, miny, maxy)

  self:computeCamPosition(id)
end

-- UPDATE and MOVE functions
-- Move and update the camera system

function CameraSystem:update(dt)
  for i,v in ipairs(self.views.list) do
    self:followActor(i)
  end
end

function CameraSystem:moveView(id, x, y)
  self.views.list[id].pos.x = x
  self.views.list[id].pos.y = y

  self:computeCamPosition(id)
  self:limitView(id)
end

function CameraSystem:computeCamPosition(id)
  local decalx = self.views.list[id].pos.onScreen.x
  local decaly = self.views.list[id].pos.onScreen.y

  local realx  = self.views.list[id].pos.x
  local realy  = self.views.list[id].pos.y

  self.views.list[id].cam.x = realx - decalx
  self.views.list[id].cam.y = realy + decaly
  -- FIXME: this workaround certainly will cause some problem but that's the only
  -- solution we seem to have right now
  -- We invert the y decalage for the camera in order to work around a problem
  -- that invert the y between it and the clipping.
end

function CameraSystem:followActor(id)
  local view = self:getView(id)

  if view.target ~= nil then
    local x, y = view.target:getViewCenter()
    x = math.floor(x)
    y = math.floor(y)
    self:moveView(id, x, y)
  end
end

-- DRAW FUNCTIONS
-- Basic callback to draw stuff

function CameraSystem:drawDebugViewBox(id)
  local viewx, viewy, vieww, viewh = self:getOnScreenViewCoordinate(id)
  utils.graphics.box(viewx, viewy, vieww, viewh)

  local xx, yy = self:getOnScreenViewCenter(id)
  love.graphics.line(xx-3, yy, xx+3, yy)
  love.graphics.line(xx, yy-3, xx, yy+3)
  local string = id .. " x:" .. viewx .. " y:" ..  viewy
  love.graphics.print(string, viewx + 4, viewy + 4)
  print(viewy)
end

function CameraSystem:drawHUD(id)
  local viewx, viewy, vieww, viewh = self:getOnScreenViewCoordinate(id)
  local view = self:getView(id)
  local string2 = id .. " (" .. viewx .. ":" .. (viewh-viewy) .. ") "


  love.graphics.setScissor(viewx, viewh-viewy, vieww, viewh)
  love.graphics.translate(viewx, viewh-viewy)
  view.target:drawHUD(id, vieww, viewh)

  love.graphics.translate(-viewx, -(viewh-viewy))
  love.graphics.setScissor( )
end

return CameraSystem
